from types import MappingProxyType
from typing import Optional, Set
from dataclasses import field
from websockets.server import WebSocketServerProtocol
from dataclasses import dataclass

from ..types import ChannelId, ClientChannel, ClientChannelId, SubscriptionId


@dataclass
class ClientState:
    """
    ClientState holds information about subscriptions from a given client, used by the server for
    bookkeeping. The `subscriptions` and `subscriptions_by_channel` are immutable, which makes them
    safe to use in concurrent (async) code without worrying that they will be mutated during
    iteration.
    """

    connection: WebSocketServerProtocol
    subscriptions: "MappingProxyType[SubscriptionId, ChannelId]" = field(
        default_factory=lambda: MappingProxyType({})
    )
    subscriptions_by_channel: "MappingProxyType[ChannelId, SubscriptionId]" = field(
        default_factory=lambda: MappingProxyType({})
    )
    advertisements_by_channel: "MappingProxyType[ClientChannelId, ClientChannel]" = (
        field(default_factory=lambda: MappingProxyType({}))
    )
    subscribed_params: Set[str] = field(default_factory=lambda: set([]))

    def remove_channel(self, removed_chan_id: ChannelId):
        sub_id = self.subscriptions_by_channel.get(removed_chan_id)
        if sub_id is not None:
            self.subscriptions = MappingProxyType(
                {
                    sub: chan
                    for sub, chan in self.subscriptions.items()
                    if chan != removed_chan_id
                }
            )
            self.subscriptions_by_channel = MappingProxyType(
                {
                    chan: subs
                    for chan, subs in self.subscriptions_by_channel.items()
                    if chan != removed_chan_id
                }
            )

    def add_subscription(self, sub_id: SubscriptionId, chan_id: ChannelId) -> bool:
        if chan_id in self.subscriptions_by_channel:
            return False

        self.subscriptions = MappingProxyType({**self.subscriptions, sub_id: chan_id})
        self.subscriptions_by_channel = MappingProxyType(
            {
                **self.subscriptions_by_channel,
                chan_id: sub_id,
            }
        )
        return True

    def remove_subscription(
        self, removed_sub_id: SubscriptionId
    ) -> Optional[ChannelId]:
        chan_id = self.subscriptions.get(removed_sub_id)
        if chan_id is None:
            return None
        self.subscriptions = MappingProxyType(
            {
                sub: chan
                for sub, chan in self.subscriptions.items()
                if sub != removed_sub_id
            }
        )
        new_subscriptions_by_channel = {
            chan: subs
            for chan, subs in self.subscriptions_by_channel.items()
            if chan != chan_id
        }
        self.subscriptions_by_channel = MappingProxyType(new_subscriptions_by_channel)

        return chan_id

    def add_client_channel(self, channel: ClientChannel) -> bool:
        if channel["id"] in self.advertisements_by_channel:
            return False

        self.advertisements_by_channel = MappingProxyType(
            {
                **self.advertisements_by_channel,
                channel["id"]: channel,
            }
        )
        return True

    def remove_client_channel(self, channel_id: ClientChannelId) -> bool:
        if channel_id not in self.advertisements_by_channel:
            return False

        self.advertisements_by_channel = MappingProxyType(
            {
                chan: subs
                for chan, subs in self.advertisements_by_channel.items()
                if chan != channel_id
            }
        )
        return True



from typing import cast
from foxglove_websocket.types import ChannelId, SubscriptionId
from websockets.server import WebSocketServerProtocol
from foxglove_websocket.server.client_state import ClientState


def test_add_subscription():
    state = ClientState(connection=cast(WebSocketServerProtocol, None))
    assert state.subscriptions == {}
    assert state.subscriptions_by_channel == {}

    state.add_subscription(SubscriptionId(0), ChannelId(100))
    assert state.subscriptions == {0: 100}
    assert state.subscriptions_by_channel == {100: 0}

    state.add_subscription(SubscriptionId(1), ChannelId(101))
    assert state.subscriptions == {0: 100, 1: 101}
    assert state.subscriptions_by_channel == {100: 0, 101: 1}

    state.add_subscription(SubscriptionId(2), ChannelId(101))
    assert state.subscriptions == {0: 100, 1: 101}
    assert state.subscriptions_by_channel == {100: 0, 101: 1}


def test_remove_subscription():
    state = ClientState(connection=cast(WebSocketServerProtocol, None))
    assert state.subscriptions == {}
    assert state.subscriptions_by_channel == {}

    state.add_subscription(SubscriptionId(0), ChannelId(100))
    state.add_subscription(SubscriptionId(1), ChannelId(101))
    state.add_subscription(SubscriptionId(2), ChannelId(102))
    assert state.subscriptions == {0: 100, 1: 101, 2: 102}
    assert state.subscriptions_by_channel == {100: 0, 101: 1, 102: 2}

    assert state.remove_subscription(SubscriptionId(99)) is None

    assert state.remove_subscription(SubscriptionId(1)) == 101
    assert state.subscriptions == {0: 100, 2: 102}
    assert state.subscriptions_by_channel == {100: 0, 102: 2}

    assert state.remove_subscription(SubscriptionId(0)) == 100
    assert state.subscriptions == {2: 102}
    assert state.subscriptions_by_channel == {102: 2}

    assert state.remove_subscription(SubscriptionId(2)) == 102
    assert state.subscriptions == {}
    assert state.subscriptions_by_channel == {}

    assert state.remove_subscription(SubscriptionId(2)) is None


def test_remove_channel():
    state = ClientState(connection=cast(WebSocketServerProtocol, None))
    assert state.subscriptions == {}
    assert state.subscriptions_by_channel == {}

    state.add_subscription(SubscriptionId(0), ChannelId(100))
    state.add_subscription(SubscriptionId(1), ChannelId(101))

    assert state.subscriptions == {0: 100, 1: 101}
    assert state.subscriptions_by_channel == {100: 0, 101: 1}

    state.remove_channel(ChannelId(999))
    assert state.subscriptions == {0: 100, 1: 101}
    assert state.subscriptions_by_channel == {100: 0, 101: 1}

    state.remove_channel(ChannelId(100))
    assert state.subscriptions == {1: 101}
    assert state.subscriptions_by_channel == {101: 1}

    state.remove_channel(ChannelId(101))
    assert state.subscriptions == {}
    assert state.subscriptions_by_channel == {}



import asyncio
import json
import logging
import pytest
from socket import AddressFamily
from struct import Struct
from typing import Dict, List, Tuple, Optional
from websockets.client import connect
from websockets.server import WebSocketServer

from foxglove_websocket.server import (
    FoxgloveServer,
    FoxgloveServerListener,
    MessageDataHeader,
)
from foxglove_websocket.types import (
    BinaryOpcode,
    ChannelId,
    ChannelWithoutId,
    ClientBinaryOpcode,
    Parameter,
    ServiceWithoutId,
    ServiceId,
)


def get_server_url(server: WebSocketServer):
    """
    Return a url that can be used to connect to the test server.
    """
    assert server.sockets
    for sock in server.sockets:
        if sock.family == AddressFamily.AF_INET:
            return f"ws://{sock.getsockname()[0]}:{sock.getsockname()[1]}"
        elif sock.family == AddressFamily.AF_INET6:  # type: ignore
            return f"ws://[{sock.getsockname()[0]}]:{sock.getsockname()[1]}"
    raise RuntimeError("Expected IPv4 or IPv6 socket")


@pytest.mark.asyncio
async def test_shutdown_before_startup():
    server = FoxgloveServer("127.0.0.1", None, "test server")
    server.start()
    server.close()
    await server.wait_closed()


@pytest.mark.asyncio
async def test_shutdown_during_startup():
    server = FoxgloveServer("127.0.0.1", None, "test server")
    server.start()
    await asyncio.sleep(0)
    server.close()
    await server.wait_closed()


@pytest.mark.asyncio
async def test_shutdown_after_startup():
    server = FoxgloveServer("127.0.0.1", None, "test server")
    server.start()
    await server.wait_opened()
    server.close()
    await server.wait_closed()


@pytest.mark.asyncio
async def test_warn_invalid_channel():
    async with FoxgloveServer("127.0.0.1", None, "test server") as server:
        async with connect(get_server_url(await server.wait_opened())) as ws:
            assert json.loads(await ws.recv())["op"] == "serverInfo"
            assert json.loads(await ws.recv())["op"] == "advertise"
            await ws.send(
                json.dumps(
                    {
                        "op": "subscribe",
                        "subscriptions": [{"id": 42, "channelId": 999}],
                    }
                )
            )
            assert json.loads(await ws.recv()) == {
                "op": "status",
                "level": 1,
                "message": "Channel 999 is not available; ignoring subscription",
            }


@pytest.mark.asyncio
async def test_disconnect_during_send(caplog: pytest.LogCaptureFixture):
    async with FoxgloveServer("127.0.0.1", None, "test server") as server:
        async with connect(get_server_url(await server.wait_opened())) as ws:
            await ws.send(
                json.dumps(
                    {
                        "op": "subscribe",
                        "subscriptions": [{"id": 42, "channelId": 999}],
                    }
                )
            )

    for record in caplog.records:
        assert record.levelno < logging.ERROR, str(record)


@pytest.mark.asyncio
async def test_listener_callbacks():
    listener_calls: List[Tuple[str, ChannelId]] = []

    class Listener(FoxgloveServerListener):
        async def on_subscribe(self, server: FoxgloveServer, channel_id: ChannelId):
            listener_calls.append(("subscribe", channel_id))

        async def on_unsubscribe(self, server: FoxgloveServer, channel_id: ChannelId):
            listener_calls.append(("unsubscribe", channel_id))

    async with FoxgloveServer("127.0.0.1", None, "test server") as server:
        server.set_listener(Listener())
        ws_server = await server.wait_opened()
        chan_id = await server.add_channel(
            {
                "topic": "t",
                "encoding": "e",
                "schemaName": "S",
                "schema": "s",
                "schemaEncoding": "s",
            }
        )
        async with connect(get_server_url(ws_server)) as ws:
            assert json.loads(await ws.recv()) == {
                "op": "serverInfo",
                "name": "test server",
                "capabilities": [],
                "metadata": None,
                "sessionId": None,
                "supportedEncodings": None,
            }
            assert json.loads(await ws.recv()) == {
                "channels": [
                    {
                        "encoding": "e",
                        "id": chan_id,
                        "schema": "s",
                        "schemaName": "S",
                        "topic": "t",
                        "schemaEncoding": "s",
                    }
                ],
                "op": "advertise",
            }
            await ws.send(
                json.dumps(
                    {
                        "op": "subscribe",
                        "subscriptions": [{"id": 42, "channelId": chan_id}],
                    }
                )
            )

    assert listener_calls == [("subscribe", chan_id), ("unsubscribe", chan_id)]


@pytest.mark.asyncio
async def test_update_channels():
    async with FoxgloveServer("127.0.0.1", None, "test server") as server:
        ws_server = await server.wait_opened()
        async with connect(get_server_url(ws_server)) as ws:
            assert json.loads(await ws.recv()) == {
                "op": "serverInfo",
                "name": "test server",
                "capabilities": [],
                "metadata": None,
                "sessionId": None,
                "supportedEncodings": None,
            }
            assert json.loads(await ws.recv()) == {
                "channels": [],
                "op": "advertise",
            }

            chan_id = await server.add_channel(
                {
                    "topic": "t",
                    "encoding": "e",
                    "schemaName": "S",
                    "schema": "s",
                    "schemaEncoding": "s",
                }
            )
            assert json.loads(await ws.recv()) == {
                "channels": [
                    {
                        "encoding": "e",
                        "id": chan_id,
                        "schema": "s",
                        "schemaName": "S",
                        "topic": "t",
                        "schemaEncoding": "s",
                    }
                ],
                "op": "advertise",
            }
            await server.remove_channel(chan_id)
            assert json.loads(await ws.recv()) == {
                "channelIds": [chan_id],
                "op": "unadvertise",
            }


@pytest.mark.asyncio
async def test_unsubscribe_during_send():
    subscribed_event = asyncio.Event()
    unsubscribed_event = asyncio.Event()

    class Listener(FoxgloveServerListener):
        async def on_subscribe(self, server: FoxgloveServer, channel_id: ChannelId):
            subscribed_event.set()

        async def on_unsubscribe(self, server: FoxgloveServer, channel_id: ChannelId):
            unsubscribed_event.set()

    async with FoxgloveServer("127.0.0.1", None, "test server") as server:
        server.set_listener(Listener())
        ws_server = await server.wait_opened()
        channel: ChannelWithoutId = {
            "topic": "t",
            "encoding": "e",
            "schemaName": "S",
            "schema": "s",
            "schemaEncoding": "s",
        }
        chan_id = await server.add_channel(channel)
        async with connect(get_server_url(ws_server)) as ws:
            assert json.loads(await ws.recv()) == {
                "op": "serverInfo",
                "name": "test server",
                "capabilities": [],
                "metadata": None,
                "sessionId": None,
                "supportedEncodings": None,
            }
            assert json.loads(await ws.recv()) == {
                "channels": [{**channel, "id": chan_id}],
                "op": "advertise",
            }

            await ws.send(
                json.dumps(
                    {
                        "op": "subscribe",
                        "subscriptions": [{"id": 42, "channelId": chan_id}],
                    }
                )
            )
            await subscribed_event.wait()

            # Force the server to yield during send_message() sending by starting a fragmented message
            assert len(ws_server.websockets) == 1
            client = next(iter(ws_server.websockets))
            fragment_future: "asyncio.Future[str]" = asyncio.Future()
            sent_first_fragment = asyncio.Event()

            async def fragments():
                yield "frag1"
                sent_first_fragment.set()
                yield await fragment_future

            pause_task = asyncio.create_task(client.send(fragments()))
            await sent_first_fragment.wait()

            payload = json.dumps({"hello": "world"}).encode()
            send_task = asyncio.create_task(server.send_message(chan_id, 100, payload))

            await ws.send(json.dumps({"op": "unsubscribe", "subscriptionIds": [42]}))
            await unsubscribed_event.wait()

            fragment_future.set_result("frag2")
            await pause_task
            await send_task

            assert await ws.recv() == "frag1frag2"
            assert (
                await ws.recv()
                == MessageDataHeader.pack(BinaryOpcode.MESSAGE_DATA, 42, 100) + payload
            )


@pytest.mark.asyncio
async def test_service_call():
    class Listener(FoxgloveServerListener):
        async def on_service_request(
            self,
            server: FoxgloveServer,
            service_id: ServiceId,
            call_id: str,
            encoding: str,
            payload: bytes,
        ) -> bytes:
            assert encoding == "json"
            assert json.loads(payload) == {"data": True}
            return json.dumps({"success": True}).encode()

    async with FoxgloveServer(
        "127.0.0.1",
        None,
        "test server",
        capabilities=["services"],
        supported_encodings=["json"],
    ) as server:
        server.set_listener(Listener())
        ws_server = await server.wait_opened()

        service: ServiceWithoutId = {
            "name": "set_bool",
            "type": "set_bool",
            "requestSchema": json.dumps(
                {
                    "type": "object",
                    "properties": {
                        "data": {"type": "boolean"},
                    },
                }
            ),
            "responseSchema": json.dumps(
                {
                    "type": "object",
                    "properties": {
                        "success": {"type": "boolean"},
                    },
                }
            ),
        }
        service_id = await server.add_service(service)

        async with connect(get_server_url(ws_server)) as ws:
            assert json.loads(await ws.recv()) == {
                "op": "serverInfo",
                "name": "test server",
                "capabilities": ["services"],
                "supportedEncodings": ["json"],
                "metadata": None,
                "sessionId": None,
            }

            advertise_msg = json.loads(await ws.recv())
            assert advertise_msg["op"] == "advertise"

            assert json.loads(await ws.recv()) == {
                "services": [{**service, "id": service_id}],
                "op": "advertiseServices",
            }

            service_header = Struct("<BIII")
            call_id = 123
            encoding = "json"
            request_header = service_header.pack(
                ClientBinaryOpcode.SERVICE_CALL_REQUEST,
                service_id,
                call_id,
                len(encoding),
            )
            request_payload = json.dumps({"data": True})
            await ws.send([request_header, encoding.encode(), request_payload.encode()])

            response = await ws.recv()
            assert isinstance(response, bytes)
            (
                op,
                res_service_id,
                res_call_id,
                encoding_length,
            ) = service_header.unpack_from(response)
            assert op == BinaryOpcode.SERVICE_CALL_RESPONSE
            assert res_service_id == service_id
            assert res_call_id == call_id
            assert encoding_length == len(encoding)
            response_payload = response[service_header.size + encoding_length :]
            assert json.loads(response_payload) == {"success": True}


@pytest.mark.asyncio
async def test_param_get_set():
    class Listener(FoxgloveServerListener):
        def __init__(self) -> None:
            self._params: Dict[str, Parameter] = {
                "int_param": Parameter(name="int_param", value=123, type=None),
                "str_param": Parameter(name="str_param", value="foo", type=None),
            }

        async def on_get_parameters(
            self,
            server: FoxgloveServer,
            param_names: List[str],
            request_id: Optional[str],
        ) -> List[Parameter]:
            if len(param_names) == 0:
                return list(self._params.values())
            else:
                return [
                    param
                    for param in self._params.values()
                    if param["name"] in param_names
                ]

        async def on_set_parameters(
            self,
            server: FoxgloveServer,
            params: List[Parameter],
            request_id: Optional[str],
        ):
            for param in params:
                self._params[param["name"]] = param
            param_names = [param["name"] for param in params]
            return [
                param for param in self._params.values() if param["name"] in param_names
            ]

    async with FoxgloveServer(
        "127.0.0.1",
        None,
        "test server",
        capabilities=["parameters"],
    ) as server:
        server.set_listener(Listener())
        ws_server = await server.wait_opened()

        async with connect(get_server_url(ws_server)) as ws:
            assert json.loads(await ws.recv()) == {
                "op": "serverInfo",
                "name": "test server",
                "capabilities": ["parameters"],
                "metadata": None,
                "sessionId": None,
                "supportedEncodings": None,
            }

            advertise_msg = json.loads(await ws.recv())
            assert advertise_msg["op"] == "advertise"

            await ws.send(json.dumps({"op": "getParameters", "parameterNames": []}))
            assert json.loads(await ws.recv()) == {
                "op": "parameterValues",
                "parameters": [
                    {"name": "int_param", "value": 123, "type": None},
                    {"name": "str_param", "value": "foo", "type": None},
                ],
                "id": None,
            }

            await ws.send(
                json.dumps({"op": "getParameters", "parameterNames": ["str_param"]})
            )
            assert json.loads(await ws.recv()) == {
                "op": "parameterValues",
                "parameters": [
                    {"name": "str_param", "value": "foo", "type": None},
                ],
                "id": None,
            }

            await ws.send(
                json.dumps(
                    {
                        "op": "setParameters",
                        "id": "set-req",
                        "parameters": [
                            {"name": "str_param", "value": "bar", "type": None},
                        ],
                    }
                )
            )
            assert json.loads(await ws.recv()) == {
                "op": "parameterValues",
                "id": "set-req",
                "parameters": [
                    {"name": "str_param", "value": "bar", "type": None},
                ],
            }



"""
This script demonstrates how to write a Foxglove WebSocket server that transmits Protobuf-encoded
data. The included Protobuf schemas are generated from https://github.com/foxglove/schemas.
"""

import asyncio
import sys
import time
from base64 import b64encode
from traceback import print_exception
from typing import Set, Type
from foxglove_websocket import run_cancellable
from foxglove_websocket.server import FoxgloveServer, FoxgloveServerListener
from foxglove_websocket.types import ChannelId

try:
    from foxglove_schemas_protobuf.SceneUpdate_pb2 import SceneUpdate
    import google.protobuf.message
    from google.protobuf.descriptor_pb2 import FileDescriptorSet
    from google.protobuf.descriptor import FileDescriptor
    from pyquaternion import Quaternion
except ImportError as err:
    print_exception(*sys.exc_info())
    print(
        "Unable to import protobuf schemas; did you forget to run `pip install 'foxglove-websocket[examples]'`?",
    )
    sys.exit(1)


def build_file_descriptor_set(
    message_class: Type[google.protobuf.message.Message],
) -> FileDescriptorSet:
    """
    Build a FileDescriptorSet representing the message class and its dependencies.
    """
    file_descriptor_set = FileDescriptorSet()
    seen_dependencies: Set[str] = set()

    def append_file_descriptor(file_descriptor: FileDescriptor):
        for dep in file_descriptor.dependencies:
            if dep.name not in seen_dependencies:
                seen_dependencies.add(dep.name)
                append_file_descriptor(dep)
        file_descriptor.CopyToProto(file_descriptor_set.file.add())  # type: ignore

    append_file_descriptor(message_class.DESCRIPTOR.file)
    return file_descriptor_set


async def main():
    class Listener(FoxgloveServerListener):
        async def on_subscribe(self, server: FoxgloveServer, channel_id: ChannelId):
            print("First client subscribed to", channel_id)

        async def on_unsubscribe(self, server: FoxgloveServer, channel_id: ChannelId):
            print("Last client unsubscribed from", channel_id)

    async with FoxgloveServer("0.0.0.0", 8765, "example server") as server:
        server.set_listener(Listener())
        chan_id = await server.add_channel(
            {
                "topic": "example_msg",
                "encoding": "protobuf",
                "schemaName": SceneUpdate.DESCRIPTOR.full_name,
                "schema": b64encode(
                    build_file_descriptor_set(SceneUpdate).SerializeToString()
                ).decode("ascii"),
                "schemaEncoding": "protobuf",
            }
        )

        i = 0
        while True:
            i += 1
            await asyncio.sleep(0.05)
            now = time.time_ns()

            scene_update = SceneUpdate()
            entity = scene_update.entities.add()
            entity.timestamp.FromNanoseconds(now)
            entity.frame_id = "root"
            cube = entity.cubes.add()
            cube.size.x = 1
            cube.size.y = 1
            cube.size.z = 1
            cube.pose.position.x = 2
            cube.pose.position.y = 0
            cube.pose.position.z = 0
            q = Quaternion(axis=[0, 0, 1], angle=i * 0.1)
            cube.pose.orientation.x = q.x
            cube.pose.orientation.y = q.y
            cube.pose.orientation.z = q.z
            cube.pose.orientation.w = q.w
            cube.color.r = 0.6
            cube.color.g = 0.2
            cube.color.b = 1
            cube.color.a = 1

            await server.send_message(chan_id, now, scene_update.SerializeToString())


if __name__ == "__main__":
    run_cancellable(main())



import asyncio
import signal
from typing import Any, Coroutine


def run_cancellable(coro: Coroutine[None, None, Any]):
    """
    Run a coroutine such that a ctrl-C interrupt will gracefully cancel its
    execution and give it a chance to clean up before returning.

    See also: https://www.roguelynn.com/words/asyncio-graceful-shutdowns/
    """
    loop = asyncio.get_event_loop()
    task = loop.create_task(coro)
    try:
        loop.add_signal_handler(signal.SIGINT, task.cancel)
    except NotImplementedError:
        # signal handlers are not available on Windows, KeyboardInterrupt will be raised instead
        pass

    try:
        try:
            loop.run_until_complete(task)
        except KeyboardInterrupt:
            task.cancel()
            loop.run_until_complete(task)
    except asyncio.CancelledError:
        pass



import asyncio
from typing import Any, Dict, List, Optional
from foxglove_websocket import run_cancellable
from foxglove_websocket.server import FoxgloveServer, FoxgloveServerListener
from foxglove_websocket.types import Parameter


async def main():
    class Listener(FoxgloveServerListener):
        def __init__(self, param_store: Dict[str, Any]) -> None:
            self._param_store = param_store

        async def on_get_parameters(
            self,
            server: FoxgloveServer,
            param_names: List[str],
            request_id: Optional[str],
        ) -> List[Parameter]:
            return [
                Parameter(name=k, value=v, type=None)
                for k, v in self._param_store.items()
                if k in param_names or len(param_names) == 0
            ]

        async def on_set_parameters(
            self,
            server: FoxgloveServer,
            params: List[Parameter],
            request_id: Optional[str],
        ):
            for param in params:
                if not param["name"].startswith("read_only"):
                    self._param_store[param["name"]] = param["value"]
            param_names = [param["name"] for param in params]
            return [
                Parameter(name=k, value=v, type=None)
                for k, v in self._param_store.items()
                if k in param_names
            ]

    async with FoxgloveServer(
        "0.0.0.0",
        8765,
        "example param server",
        capabilities=["parameters", "parametersSubscribe"],
    ) as server:
        param_store: Dict[str, Any] = {
            "int_param": 0,
            "str_param": "asdf",
            "bool_param": True,
            "int_array_param": [1, 2, 3],
            "str_array_param": ["abc", "def", "fgh"],
            "bool_array_param": [True, False, True],
            "read_only_str_param": "can't change me",
        }

        server.set_listener(Listener(param_store))

        i = 0
        while True:
            i += 1
            await asyncio.sleep(3.0)
            param_store["int_param"] = i
            await server.update_parameters(
                [Parameter(name="int_param", value=param_store["int_param"], type=None)]
            )


if __name__ == "__main__":
    run_cancellable(main())



from abc import ABC
import asyncio
import inspect
import json
import logging
from struct import Struct
from typing import (
    Any,
    Dict,
    List,
    Mapping,
    Optional,
    Set,
    Union,
    Tuple,
    cast,
    Coroutine,
)
from websockets.server import serve, WebSocketServer, WebSocketServerProtocol
from websockets.exceptions import ConnectionClosed
from websockets.typing import Data, Subprotocol

from .client_state import ClientState
from ..types import (
    BinaryOpcode,
    Channel,
    ChannelId,
    ChannelWithoutId,
    ClientBinaryOpcode,
    ClientChannel,
    ClientChannelId,
    ClientJsonMessage,
    Parameter,
    Service,
    ServiceId,
    ServiceWithoutId,
    SubscriptionId,
    ServerJsonMessage,
    StatusLevel,
)


def _get_default_logger():
    logger = logging.getLogger("FoxgloveServer")
    logger.setLevel(logging.DEBUG)
    handler = logging.StreamHandler()
    handler.setFormatter(logging.Formatter("%(asctime)s: [%(levelname)s] %(message)s"))
    logger.addHandler(handler)
    return logger


MessageDataHeader = Struct("<BIQ")
TimeDataHeader = Struct("<BQ")
ClientMessageHeader = Struct("<BI")
ServiceCallRequestHeader = Struct("<BIII")
ServiceCallResponseHeader = ServiceCallRequestHeader


class FoxgloveServerListener(ABC):
    def on_subscribe(
        self, server: "FoxgloveServer", channel_id: ChannelId
    ) -> Union[None, Coroutine[Any, Any, None]]:
        """
        Called when the first client subscribes to `channel_id`.
        """
        ...

    def on_unsubscribe(
        self, server: "FoxgloveServer", channel_id: ChannelId
    ) -> Union[None, Coroutine[Any, Any, None]]:
        """
        Called when the last subscribed client unsubscribes from `channel_id`.
        """
        ...

    async def on_client_advertise(
        self, server: "FoxgloveServer", channel: ClientChannel
    ):
        """
        Called when a new client channel is advertised.
        """
        ...

    async def on_client_unadvertise(
        self, server: "FoxgloveServer", channel_id: ClientChannelId
    ):
        """
        Called when a client channel is unadvertised.
        """
        ...

    async def on_client_message(
        self, server: "FoxgloveServer", channel_id: ClientChannelId, payload: bytes
    ):
        """
        Called when a client channel message is received.
        """
        ...

    async def on_service_request(
        self,
        server: "FoxgloveServer",
        service_id: ServiceId,
        call_id: str,
        encoding: str,
        payload: bytes,
    ) -> bytes:
        """
        Called when a service request is made.
        """
        ...

    async def on_get_parameters(
        self,
        server: "FoxgloveServer",
        param_names: List[str],
        request_id: Optional[str],
    ) -> List[Parameter]:
        """
        Called when parameter values are requested.
        """
        ...

    async def on_set_parameters(
        self,
        server: "FoxgloveServer",
        params: List[Parameter],
        request_id: Optional[str],
    ) -> List[Parameter]:
        """
        Called when parameteres are to be modified.
        """
        ...

    async def on_parameters_subscribe(
        self, server: "FoxgloveServer", param_name: List[str], subscribe: bool
    ):
        """
        Called when parameters are subscribed/unsubscribed.
        """
        ...


class FoxgloveServer:
    _clients: Tuple[ClientState, ...]
    _channels: Dict[ChannelId, Channel]
    _next_channel_id: ChannelId
    _services: Dict[ServiceId, Service]
    _next_service_id: ServiceId
    _subscribed_params: Set[str]
    _logger: logging.Logger
    _listener: Optional[FoxgloveServerListener]
    _opened: "asyncio.Future[WebSocketServer]"

    def __init__(
        self,
        host: str,
        port: Optional[int],
        name: str,
        *,
        capabilities: List[str] = [],
        supported_encodings: Optional[List[str]] = None,
        metadata: Optional[Mapping[str, str]] = None,
        session_id: Optional[str] = None,
        logger: logging.Logger = _get_default_logger(),
        server_kwargs: Dict[str, Any] = {"compression": None},
    ):
        self.host = host
        self.port = port
        self.name = name
        self.capabilities = capabilities
        self.supported_encodings = supported_encodings
        self.metadata = metadata
        self.session_id = session_id
        self._clients = ()
        self._channels = {}
        self._next_channel_id = ChannelId(0)
        self._services = {}
        self._next_service_id = ServiceId(0)
        self._subscribed_params = set([])
        self._logger = logger
        self._server_kwargs = server_kwargs
        self._listener = None
        self._opened = asyncio.get_running_loop().create_future()

    def set_listener(self, listener: FoxgloveServerListener):
        self._listener = listener

    def _any_subscribed(self, chan_id: ChannelId):
        return any(
            chan_id in client.subscriptions_by_channel for client in self._clients
        )

    async def __aenter__(self):
        self.start()
        return self

    async def __aexit__(self, exc_type: Any, exc_val: Any, traceback: Any):
        self.close()
        await self.wait_closed()

    async def wait_opened(self):
        return await self._opened

    def start(self):
        self._task = asyncio.create_task(self._run())

    def close(self):
        self._logger.info("Shutting down...")
        self._task.cancel()

    async def wait_closed(self):
        try:
            await self._task
        except asyncio.CancelledError:
            pass

    async def _run(self):
        # TODO: guard against multiple run calls?
        self._logger.info("Starting server...")
        try:
            server = await serve(
                self._handle_connection,
                self.host,
                self.port,
                subprotocols=[Subprotocol("foxglove.websocket.v1")],
                **self._server_kwargs,
            )
            self._opened.set_result(server)
        except asyncio.CancelledError:
            self._logger.info("Canceled during server startup")
            return

        for sock in server.sockets or []:
            self._logger.info("Server listening on %s", sock.getsockname())
        try:
            await server.wait_closed()
        except asyncio.CancelledError:
            server.close()
            await server.wait_closed()
            self._logger.info("Server closed")

    async def add_channel(self, channel: ChannelWithoutId):
        new_id = self._next_channel_id
        self._next_channel_id = ChannelId(new_id + 1)
        new_channel = Channel(id=new_id, **channel)
        self._channels[new_id] = new_channel
        # Any clients added during await will already see the new channel.
        for client in self._clients:
            await self._send_json(
                client.connection,
                {
                    "op": "advertise",
                    "channels": [new_channel],
                },
            )
        return new_id

    async def remove_channel(self, chan_id: ChannelId):
        del self._channels[chan_id]
        # Any clients added during await will not have received info about the new channel.
        for client in self._clients:
            client.remove_channel(chan_id)

            await self._send_json(
                client.connection,
                {
                    "op": "unadvertise",
                    "channelIds": [chan_id],
                },
            )

    async def add_service(self, service: ServiceWithoutId) -> ServiceId:
        new_id = self._next_service_id
        self._next_service_id = ServiceId(new_id + 1)
        new_service = Service(id=new_id, **service)
        self._services[new_id] = new_service
        # Any clients added during await will already see the new service.
        for client in self._clients:
            await self._send_json(
                client.connection,
                {
                    "op": "advertiseServices",
                    "services": [new_service],
                },
            )
        return new_id

    async def remove_service(self, service_id: ServiceId):
        del self._services[service_id]
        # Any clients added during await will not have received info about the new service.
        for client in self._clients:
            await self._send_json(
                client.connection,
                {
                    "op": "unadvertiseServices",
                    "serviceIds": [service_id],
                },
            )

    async def update_parameters(self, parameters: List[Parameter]):
        for client in self._clients:
            params_of_interest = [
                p for p in parameters if p["name"] in client.subscribed_params
            ]
            if len(params_of_interest):
                await self._send_json(
                    client.connection,
                    {
                        "op": "parameterValues",
                        "parameters": params_of_interest,
                        "id": None,
                    },
                )

    async def send_message(self, chan_id: ChannelId, timestamp: int, payload: bytes):
        for client in self._clients:
            sub_id = client.subscriptions_by_channel.get(chan_id, None)
            if sub_id is not None:
                await self._send_message_data(
                    client.connection,
                    subscription=sub_id,
                    timestamp=timestamp,
                    payload=payload,
                )

    async def reset_session_id(self, newSessionId: Optional[str] = None):
        """
        Reset session Id and send new server info to clients.
        """
        self.session_id = newSessionId
        for client in self._clients:
            await self._send_server_info(client.connection)

    async def broadcast_time(self, timestamp: int):
        msg = TimeDataHeader.pack(BinaryOpcode.TIME, timestamp)
        for client in self._clients:
            try:
                await client.connection.send(msg)
            except ConnectionClosed:
                pass

    async def _send_json(
        self, connection: WebSocketServerProtocol, msg: ServerJsonMessage
    ):
        try:
            await connection.send(json.dumps(msg, separators=(",", ":")))
        except ConnectionClosed:
            pass

    async def _send_message_data(
        self,
        connection: WebSocketServerProtocol,
        *,
        subscription: SubscriptionId,
        timestamp: int,
        payload: bytes,
    ):
        try:
            header = MessageDataHeader.pack(
                BinaryOpcode.MESSAGE_DATA, subscription, timestamp
            )
            await connection.send([header, payload])
        except ConnectionClosed:
            pass

    async def _send_server_info(self, connection: WebSocketServerProtocol) -> None:
        await self._send_json(
            connection,
            {
                "op": "serverInfo",
                "name": self.name,
                "capabilities": self.capabilities,
                "supportedEncodings": self.supported_encodings,
                "metadata": self.metadata,
                "sessionId": self.session_id,
            },
        )

    async def _handle_connection(
        self, connection: WebSocketServerProtocol, path: str
    ) -> None:
        self._logger.info(
            "Connection to %s opened via %s", connection.remote_address, path
        )

        client = ClientState(connection=connection)
        self._clients += (client,)

        try:
            await self._send_server_info(connection)
            await self._send_json(
                connection,
                {
                    "op": "advertise",
                    "channels": list(self._channels.values()),
                },
            )
            if "services" in self.capabilities:
                await self._send_json(
                    connection,
                    {
                        "op": "advertiseServices",
                        "services": list(self._services.values()),
                    },
                )
            async for raw_message in connection:
                await self._handle_raw_client_message(client, raw_message)

        except ConnectionClosed as closed:
            self._logger.info(
                "Connection to %s closed: %s %r",
                connection.remote_address,
                closed.code,
                closed.reason,
            )

        except Exception:
            self._logger.exception(
                "Error handling client connection %s", connection.remote_address
            )
            await connection.close(1011)  # Internal Error

        finally:
            potential_unsubscribes = client.subscriptions_by_channel.keys()
            self._clients = tuple(c for c in self._clients if c != client)
            if self._listener:
                for chan_id in potential_unsubscribes:
                    if not self._any_subscribed(chan_id):
                        result = self._listener.on_unsubscribe(self, chan_id)
                        if inspect.isawaitable(result):
                            await result

    async def _send_status(
        self, connection: WebSocketServerProtocol, level: StatusLevel, msg: str
    ) -> None:
        await self._send_json(
            connection,
            {
                "op": "status",
                "level": level,
                "message": msg,
            },
        )

    async def _handle_raw_client_message(self, client: ClientState, raw_message: Data):
        try:
            if isinstance(raw_message, str):
                message = json.loads(raw_message)
                self._logger.debug("Got message: %s", message)
                if not isinstance(message, dict):
                    raise TypeError(f"Expected JSON object, got {type(message)}")
                await self._handle_client_text_message(
                    client, cast(ClientJsonMessage, message)
                )
            else:
                await self._handle_client_binary_message(client, raw_message)
        except ConnectionClosed:
            pass
        except Exception as exc:
            self._logger.exception("Error handling message %s", raw_message)
            await self._send_status(
                client.connection, StatusLevel.ERROR, f"{type(exc).__name__}: {exc}"
            )

    async def _handle_client_text_message(
        self, client: ClientState, message: ClientJsonMessage
    ):
        if message["op"] == "subscribe":
            for sub in message["subscriptions"]:
                chan_id = sub["channelId"]
                sub_id = sub["id"]
                if sub_id in client.subscriptions:
                    await self._send_status(
                        client.connection,
                        StatusLevel.ERROR,
                        f"Client subscription id {sub_id} was already used; ignoring subscription",
                    )
                    continue
                chan = self._channels.get(chan_id)
                if chan is None:
                    await self._send_status(
                        client.connection,
                        StatusLevel.WARNING,
                        f"Channel {chan_id} is not available; ignoring subscription",
                    )
                    continue
                first_subscription = not self._any_subscribed(chan_id)
                if client.add_subscription(sub_id, chan_id):
                    self._logger.debug(
                        "Client %s subscribed to channel %s",
                        client.connection.remote_address,
                        chan_id,
                    )
                    if self._listener and first_subscription:
                        result = self._listener.on_subscribe(self, chan_id)
                        if inspect.isawaitable(result):
                            await result
                else:
                    await self._send_status(
                        client.connection,
                        StatusLevel.WARNING,
                        f"Client is already subscribed to channel {chan_id}; ignoring subscription",
                    )
                    continue

        elif message["op"] == "unsubscribe":
            for sub_id in message["subscriptionIds"]:
                chan_id = client.remove_subscription(sub_id)
                if chan_id is None:
                    await self._send_status(
                        client.connection,
                        StatusLevel.WARNING,
                        f"Client subscription id {sub_id} did not exist; ignoring unsubscription",
                    )
                    continue
                self._logger.debug(
                    "Client %s unsubscribed from channel %s",
                    client.connection.remote_address,
                    chan_id,
                )
                if self._listener and not self._any_subscribed(chan_id):
                    result = self._listener.on_unsubscribe(self, chan_id)
                    if inspect.isawaitable(result):
                        await result
        elif message["op"] == "advertise":
            for channel in message["channels"]:
                if not client.add_client_channel(channel):
                    self._logger.error(f"Failed to add client channel {channel['id']}")
                    await self._send_status(
                        client.connection,
                        StatusLevel.WARNING,
                        f"Failed to add client channel {channel['id']}",
                    )
                    continue
                self._logger.debug(
                    "Client %s advertised channel %d (%s)",
                    client.connection.remote_address,
                    channel["id"],
                    channel["topic"],
                )
                if self._listener:
                    await self._listener.on_client_advertise(self, channel)
        elif message["op"] == "unadvertise":
            for channel_id in message["channelIds"]:
                if not client.remove_client_channel(channel_id):
                    self._logger.error(f"Failed to remove client channel {channel_id}")
                    await self._send_status(
                        client.connection,
                        StatusLevel.WARNING,
                        f"Failed to remove client channel {channel_id}",
                    )
                    continue
                self._logger.debug(
                    "Client %s unadvertised channel %d",
                    client.connection.remote_address,
                    channel_id,
                )
                if self._listener:
                    await self._listener.on_client_unadvertise(self, channel_id)
        elif message["op"] == "getParameters":
            if self._listener:
                request_id = message.get("id", None)
                params = await self._listener.on_get_parameters(
                    self, message["parameterNames"], request_id
                )
                await self._send_json(
                    client.connection,
                    {
                        "op": "parameterValues",
                        "parameters": params,
                        "id": request_id,
                    },
                )
        elif message["op"] == "setParameters":
            if self._listener:
                request_id = message.get("id", None)
                updated_params = await self._listener.on_set_parameters(
                    self, message["parameters"], request_id
                )
                if request_id is not None:
                    await self._send_json(
                        client.connection,
                        {
                            "op": "parameterValues",
                            "parameters": updated_params,
                            "id": request_id,
                        },
                    )
                await self.update_parameters(updated_params)
        elif message["op"] == "subscribeParameterUpdates":
            new_param_subscriptions = [
                name
                for name in message["parameterNames"]
                if name not in self._subscribed_params
            ]
            client.subscribed_params.update(message["parameterNames"])
            self._subscribed_params.update(new_param_subscriptions)
            if self._listener:
                await self._listener.on_parameters_subscribe(
                    self, new_param_subscriptions, True
                )
        elif message["op"] == "unsubscribeParameterUpdates":
            new_param_unsubscriptions = [
                name
                for name in message["parameterNames"]
                if name in self._subscribed_params
            ]
            client.subscribed_params -= set(message["parameterNames"])
            self._subscribed_params -= set(new_param_unsubscriptions)
            if self._listener:
                await self._listener.on_parameters_subscribe(
                    self, new_param_unsubscriptions, False
                )
        else:
            raise ValueError(f"Unrecognized client opcode: {message['op']}")

    async def _handle_client_binary_message(
        self, client: ClientState, message: bytes
    ) -> None:
        if len(message) < 5:
            msg = f"Received invalid binary message of size {len(message)}"
            self._logger.error(msg)
            await self._send_status(
                client.connection,
                StatusLevel.ERROR,
                msg,
            )
            return

        op = message[0]

        if op == ClientBinaryOpcode.MESSAGE_DATA:
            _, channel_id = ClientMessageHeader.unpack_from(message)
            payload = message[ClientMessageHeader.size :]

            if not channel_id in client.advertisements_by_channel:
                msg = f"Channel {channel_id} not registered by client {client.connection.remote_address}"
                self._logger.error(msg)
                await self._send_status(
                    client.connection,
                    StatusLevel.ERROR,
                    msg,
                )
                return

            if self._listener:
                await self._listener.on_client_message(self, channel_id, payload)
        elif op == ClientBinaryOpcode.SERVICE_CALL_REQUEST:
            (
                _,
                service_id,
                call_id,
                encoding_length,
            ) = ServiceCallRequestHeader.unpack_from(message)
            service = self._services.get(service_id)
            if service is None:
                msg = f"Unknown service {service_id}"
                self._logger.error(msg)
                await self._send_status(
                    client.connection,
                    StatusLevel.ERROR,
                    msg,
                )
                return

            offset = ServiceCallRequestHeader.size
            encoding = message[offset : offset + encoding_length]
            offset = ServiceCallRequestHeader.size + encoding_length
            payload = message[offset:]
            if self._listener:
                response = await self._listener.on_service_request(
                    self, service_id, call_id, encoding.decode(), payload
                )
                try:
                    header = ServiceCallResponseHeader.pack(
                        BinaryOpcode.SERVICE_CALL_RESPONSE,
                        service_id,
                        call_id,
                        encoding_length,
                    )
                    await client.connection.send([header, encoding, response])
                except ConnectionClosed:
                    pass

        else:
            msg = f"Received binary message with invalid operation {op}"
            self._logger.error(msg)
            await self._send_status(
                client.connection,
                StatusLevel.ERROR,
                msg,
            )



import asyncio
import json
import time
from foxglove_websocket import run_cancellable
from foxglove_websocket.server import FoxgloveServer, FoxgloveServerListener
from foxglove_websocket.types import (
    ChannelId,
    ClientChannel,
    ClientChannelId,
    ServiceId,
)


async def main():
    class Listener(FoxgloveServerListener):
        async def on_subscribe(self, server: FoxgloveServer, channel_id: ChannelId):
            print("First client subscribed to", channel_id)

        async def on_unsubscribe(self, server: FoxgloveServer, channel_id: ChannelId):
            print("Last client unsubscribed from", channel_id)

        async def on_client_advertise(
            self, server: FoxgloveServer, channel: ClientChannel
        ):
            print("Client advertise:", json.dumps(channel))

        async def on_client_unadvertise(
            self, server: FoxgloveServer, channel_id: ClientChannelId
        ):
            print("Client unadvertise:", channel_id)

        async def on_client_message(
            self, server: FoxgloveServer, channel_id: ClientChannelId, payload: bytes
        ):
            msg = json.loads(payload)
            print(f"Client message on channel {channel_id}: {msg}")

        async def on_service_request(
            self,
            server: FoxgloveServer,
            service_id: ServiceId,
            call_id: str,
            encoding: str,
            payload: bytes,
        ) -> bytes:
            if encoding != "json":
                return json.dumps(
                    {"success": False, "error": f"Invalid encoding {encoding}"}
                ).encode()

            request = json.loads(payload)
            if "data" not in request:
                return json.dumps(
                    {"success": False, "error": f"Missing key 'data'"}
                ).encode()

            print(f"Service request on service {service_id}: {request}")
            return json.dumps(
                {"success": True, "message": f"Received boolean: {request['data']}"}
            ).encode()

    async with FoxgloveServer(
        "0.0.0.0",
        8765,
        "example server",
        capabilities=["clientPublish", "services"],
        supported_encodings=["json"],
    ) as server:
        server.set_listener(Listener())
        chan_id = await server.add_channel(
            {
                "topic": "example_msg",
                "encoding": "json",
                "schemaName": "ExampleMsg",
                "schema": json.dumps(
                    {
                        "type": "object",
                        "properties": {
                            "msg": {"type": "string"},
                            "count": {"type": "number"},
                        },
                    }
                ),
                "schemaEncoding": "jsonschema",
            }
        )
        await server.add_service(
            {
                "name": "example_set_bool",
                "requestSchema": json.dumps(
                    {
                        "type": "object",
                        "properties": {
                            "data": {"type": "boolean"},
                        },
                    }
                ),
                "responseSchema": json.dumps(
                    {
                        "type": "object",
                        "properties": {
                            "success": {"type": "boolean"},
                            "message": {"type": "string"},
                        },
                    }
                ),
                "type": "example_set_bool",
            }
        )

        i = 0
        while True:
            i += 1
            await asyncio.sleep(0.2)
            await server.send_message(
                chan_id,
                time.time_ns(),
                json.dumps({"msg": "Hello!", "count": i}).encode("utf8"),
            )


if __name__ == "__main__":
    run_cancellable(main())



from typing import (
    List,
    Literal,
    Mapping,
    NewType,
    Optional,
    TypedDict,
    Union,
)
from enum import IntEnum

ChannelId = NewType("ChannelId", int)
SubscriptionId = NewType("SubscriptionId", int)
ClientChannelId = NewType("ClientChannelId", int)
ServiceId = NewType("ServiceId", int)


class Parameter(TypedDict):
    name: str
    value: Union[
        int,
        float,
        bool,
        str,
        List[int],
        List[float],
        List[bool],
        List[str],
    ]
    type: Optional[str]


class Subscription(TypedDict):
    id: SubscriptionId
    channelId: ChannelId


class ClientChannel(TypedDict):
    id: ClientChannelId
    topic: str
    encoding: str
    schemaName: str


class Subscribe(TypedDict):
    op: Literal["subscribe"]
    subscriptions: List[Subscription]


class Unsubscribe(TypedDict):
    op: Literal["unsubscribe"]
    subscriptionIds: List[SubscriptionId]


class ClientAdvertise(TypedDict):
    op: Literal["advertise"]
    channels: List[ClientChannel]


class ClientUnadvertise(TypedDict):
    op: Literal["unadvertise"]
    channelIds: List[ClientChannelId]


class GetParameters(TypedDict):
    op: Literal["getParameters"]
    parameterNames: List[str]
    id: Optional[str]


class SetParameters(TypedDict):
    op: Literal["setParameters"]
    parameters: List[Parameter]
    id: Optional[str]


class SubscribeParameterUpdates(TypedDict):
    op: Literal["subscribeParameterUpdates"]
    parameterNames: List[str]


class UnsubscribeParameterUpdates(TypedDict):
    op: Literal["unsubscribeParameterUpdates"]
    parameterNames: List[str]


ClientJsonMessage = Union[
    Subscribe,
    Unsubscribe,
    ClientAdvertise,
    ClientUnadvertise,
    GetParameters,
    SetParameters,
    SubscribeParameterUpdates,
    UnsubscribeParameterUpdates,
]


class BinaryOpcode(IntEnum):
    MESSAGE_DATA = 1
    TIME = 2
    SERVICE_CALL_RESPONSE = 3


class ClientBinaryOpcode(IntEnum):
    MESSAGE_DATA = 1
    SERVICE_CALL_REQUEST = 2


class ServerInfo(TypedDict):
    op: Literal["serverInfo"]
    name: str
    capabilities: List[str]
    supportedEncodings: Optional[List[str]]
    metadata: Optional[Mapping[str, str]]
    sessionId: Optional[str]


class StatusLevel(IntEnum):
    INFO = 0
    WARNING = 1
    ERROR = 2


class StatusMessage(TypedDict):
    op: Literal["status"]
    level: StatusLevel
    message: str


class ChannelWithoutId(TypedDict):
    topic: str
    encoding: str
    schemaName: str
    schema: str
    schemaEncoding: Optional[str]


class Channel(ChannelWithoutId):
    id: ChannelId


class ServiceWithoutId(TypedDict):
    name: str
    type: str
    requestSchema: str
    responseSchema: str


class Service(ServiceWithoutId):
    id: ServiceId


class Advertise(TypedDict):
    op: Literal["advertise"]
    channels: List[Channel]


class Unadvertise(TypedDict):
    op: Literal["unadvertise"]
    channelIds: List[ChannelId]


class AdvertiseServices(TypedDict):
    op: Literal["advertiseServices"]
    services: List[Service]


class UnadvertiseServices(TypedDict):
    op: Literal["unadvertiseServices"]
    serviceIds: List[ServiceId]


class ParameterValues(TypedDict):
    op: Literal["parameterValues"]
    parameters: List[Parameter]
    id: Optional[str]


ServerJsonMessage = Union[
    ServerInfo,
    StatusMessage,
    Advertise,
    Unadvertise,
    AdvertiseServices,
    UnadvertiseServices,
    ParameterValues,
]
